! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

*+CALDB_INFO
        subroutine caldb_info(chatter, mode,missn, inst, ierr)

        use param

        IMPLICIT NONE
        character*(*) missn,inst
        character*(*) mode
        integer chatter, ierr
c 
c Description:
c  The subroutine checks that the CALDB is accessible to the user. If 
c NOT, then an error message is written to STDERR (irrespective of chatter)
c and IERR is returned as non-zero. If a CALDB is accessible, then depending 
c upon the chatter & mode flags, various levels of information are written 
c to STDOUT (nothing if chatter=0.0)
c 
c Passed parameters
c  CHATTER       i   : (int) Chattiness flag (0 for silent running)
c  MODE          i   : (int) Mode flag (for future use)
c  MISSN         i   : (str) Mission string
c  INSTR         i   : (str) Instrument name
c  IERR            r : (int) Error flag (zero, if CALDB appears set-up)
c
c Called Routines:
c  subroutine CTRLOG     : (CALLIB) Translates system logical/envar
c  subroutine WTINFO     : (CALLIB) Writes callib/roslib info to STDOUT
c  subroutine WTERRM     : (CALLIB) Writes callib/roslib error message
c
c Compilation & Linking
c  link with FITSIO & CALLIB & FTOOLS
c
c Origin:
c  Original
c
c Authors/Modification History:
c  Ian M George     (1.0.0: 1995 Nov 29) original
c  Ian M George     (1.1.0: 1996 Feb 06) also checks for CALDBALIAS
c  Lorraine Breedon (1.2.0: 1997 Aug 15) gets rid of instrument 'alias' problem
C                                        when calling rdcnfg routine
c  Peter D Wilson   (1.2.1: 1998 Jul 20) restore ascaarf mods (from Ian/1996),
c                                        hiding messages if caldb not setup
c                                        and chatter<0
c        character(7) version
c        parameter (version = '1.2.1')
*- 
c Internals 
        character(10) subname
        parameter (subname = 'caldb_info')
        integer caldblen, cnfglen, aliaslen, fcstln
        character(20)  cnfgvar, caldbvar, aliasvar
        character(160) message
        character(120) cnfgpth,caldbpth
        character(120) aliaspth
        character(120) ciffil, datadir
        logical quiet

c Initialize
        caldbvar = 'CALDB'
        cnfgvar = 'CALDBCONFIG'
        aliasvar = 'CALDBALIAS'
        ierr = 0


c Give user info if requested
c         message = ' using '//subname//' '//version
c         call wtinfo(chatter,20,1,message)

c See if the user has set the environment variable specified in the
c CALDBVAR argument
        call ctrlog(caldbvar,fcstln(caldbvar),caldbpth,caldblen)
        if (caldblen .eq. 0) then
c           if (chatter.ge.0) call wterrm(subname, version,
c     &          'CALDB environment variable not set')
           call wrerr(subname, 2, 'CALDB environment variable not set')
           message = 'The environ-var/logical CALDB must'//
     &          ' be set to point to the top of the Caldb'//
     &          ' directory structure'
c           call wtinfo(chatter,5,1, message)
           call wrlog(subname, message)
           message = 'See the Caldb Users Guide (CAL/GEN/94-002)'//
     &          ' for details'
c           call wtinfo(chatter,5,2,message)
           call wrlog(subname, message)
           ierr = 2
           goto 999
        else
c           call wtinfo(chatter,10,2,
c     &          'environ-var/logical CALDB defined')
           call wrchatty(subname, 'Environ-var/logical CALDB is defined.')
           message = '  CALDB path = '//caldbpth(:caldblen)
c           call wtinfo(chatter,15,3,message)
           call wrchatty(subname, message)
        endif
c See if the user has set the environment variable specified in the
c CNFGVAR argument
        call ctrlog(cnfgvar,fcstln(cnfgvar),cnfgpth,cnfglen)
        if (cnfglen .eq. 0) then
c           call wterrm(subname, version,
c     &          'CALDBCONFIG environ-var/logical not set')
           call wrerr(subname, 2, 'CALDBCONFIG environment variable not set')
           message = 'The environment variable/logical CALDBCONFIG '//
     &          ' must be set to point your local Caldb configuration'//
     &          ' file'
c           call wtinfo(chatter,5,1, message)
           call wrlog(subname, message)
           message = 'See the Caldb Users Guide (CAL/GEN/94-002)'//
     &          ' for details'
c           call wtinfo(chatter,5,2,message)
           call wrlog(subname, message)
           ierr = 2
           goto 999
        else
c           call wtinfo(chatter,10,2,
c     &          'environ-var/logical CALDBCONFIG defined')
           call wrchatty(subname, 'Environ-var/logical CALDBCONFIG is defined.')
           message = '  CALDBCONFIG file = '//cnfgpth(:cnfglen)
c           call wtinfo(chatter,15,3,message)
           call wrchatty(subname, message)
        endif

c See if the user has set the environment variable specified in the
c ALIASVAR argument
        call ctrlog(aliasvar,fcstln(aliasvar),aliaspth,aliaslen)
        if (aliaslen .eq. 0) then
c           call wterrm(subname, version,
c     &          'CALDBALIAS environ-var/logical not set')
           call wrerr(subname, 2, 'CALDBALIAS environment variable not set')
           message = 'The environment variable/logical CALDBALIAS '//
     &          ' must be set to point your local Caldb alias'//
     &          ' file'
c           call wtinfo(chatter,5,1, message)
           call wrlog(subname, message)
           message = 'See the Caldb Users Guide (CAL/GEN/94-002)'//
     &          ' for details'
c           call wtinfo(chatter,5,2,message)
           call wrlog(subname, message)
           ierr = 2
           goto 999
        else
c           call wtinfo(chatter,10,2,
c     &          'environ-var/logical CALDBALIAS defined')
           call wrchatty(subname, 'Environ-var/logical CALDBALIAS is defined.')
           message = '  CALDBALIAS file = '//aliaspth(:aliaslen)
c           call wtinfo(chatter,15,3,message)
           call wrchatty(subname, message)
        endif

c Now perform mode-specific checks....
        if(mode.eq.'INST') then
          if (inst(:fcstln(inst)).eq.'xrt'.and.missn.eq.'ASCA') then
              inst='xrt1'
          elseif (inst(:fcstln(inst)).eq.'sis'.and.missn.eq.'ASCA') then
               inst='sis0'
          elseif (inst(:fcstln(inst)).eq.'gis'.and.missn.eq.'ASCA') then
             inst='gis2'
          elseif (inst(:fcstln(inst)).eq.'pspc'.and.missn.eq.'ROSAT')
     &       then
             inst='pspcb'
          endif
   
           quiet = .true.
           if(chatter.gt.20) quiet = .false.
           call rdcnfg(missn,inst,quiet,ciffil,datadir,ierr)
           if(ierr.ne.0) then
              message = 'CALDB NOT correctly configured for the '//
     &             inst(:fcstln(inst))//' instrument onboard '//
     &             missn(:fcstln(missn))
c              call wterrm(subname, version,message)
              call wrerr(subname, 2, message)
              goto 999
           else
              message = 'CALDB is configured for the '//
     &             inst(:fcstln(inst))//' instrument onboard '//
     &             missn(:fcstln(missn))
c              call wtinfo(chatter,10,2,message)
              call wrchatty(subname, message)
              message = 'Cal Index File: '//ciffil(:fcstln(ciffil))
c              call wtinfo(chatter,15,3,message)
              call wrchatty(subname, message)
               message = 'Data directory: '//datadir(:fcstln(datadir))
c              call wtinfo(chatter,15,3,message)
              call wrchatty(subname, message)
           endif


        endif


999     if(ierr.ne.0) then
c           if (chatter.ge.0) call wterrm(subname, version, 
c     &          'Serious problem with CALDB set-up')
          call wrerr(subname, 3, 'Serious problem with CALDB set-up.')
        endif

        return
        end
