! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

*+ERO_CIFSL
        subroutine ero_cifsl(chatter,cif,tele,instr,detnam,filt,codenam,
     &  startreftime, stopreftime, expr,maxret,file,extno,online,
     &  nfound,status)

C eRO: This is a version of the cifsl2 routine modified by HB:
C      now reads calvers file given in $SASS_CALVERS environment variable (full path);
C      only outputs calibration files which have the version number specified in the 
C      calvers file. If no calvers file is present or no version number for the requested
C      calibration code and instrument is found, the calibration file with the highest
C      calibration file  will be returned. If an override file name is specified in the
C      calvers file, it will be returned instead of the caldb file. Contrary to the
C      unmodifed routine, exactly one calibration file name will be returned (nfound=1).
c
c      2020-01-20 IMS: The present routine may now (again) return >1 file name. The
c      filter process consists of 3 screens. The first screen filters out all file names
c      except those which match instrument, filter etc. - indeed all criteria except
c      time and version. The second screen further filters candidate names on the basis
c      of time; the third screen considers version (supposing no match to a version
c      number was found in the $SASS_CALVERS file). Screen 3 has been modified; 1 and 2
c      are left as they were.
c
c      Screen 3 becomes active if there are more than 1 file passing screen 2 that have
c      the same value of REF_TIME. In this case only that file with the highest version
c      number will be returned.
     
C eRO: use statements
        use param
        use iso_fortran_env

        implicit none
        character*(*) cif, tele, instr, codenam, expr
        double Precision startreftime, stopreftime
        integer       maxret, nfound, chatter
        character*(*) detnam,filt
        character*(*) file(maxret),online(maxret)
        integer       extno(maxret), status
        
C
C Description: 
C  Opens the Calibration Index File specified in the CIF argument, and 
C searches the TELESCOP, INSTRUME, DETNAM, FILTER, and CAL_CNAM columns for 
C values matching those found in TELE, INSTR, DETNAM, FILT, and CODENAM 
C passed parameters.  If the DETNAM or FILT argument is passed as a '-' then
C the DETNAM and FILTER columns are not searched.  A value of 'NONE' in the 
C DETNAM or FILTER column will match any value in DETNAM or FILT.  
C The REF_TIME column is searched for the largest value which is also less 
C than the STARTREFTIME argument.  The EXPR string is parsed and formed into 
C Calibration boundary strings which are then compared to the CAL_CBD column 
C values.  If EXPR is '-', then the CAL_CBD column is not searched.
C
C When a row in the CIF is found that meets the above criteria, the values of 
C the CALDB environment variable (logical) the CAL_DIR and CAL_FILE columns
C are formed into a system dependent path.  This path points at the 
C calibration file having the selection characteristics specified above, and 
C is returned in the FILE argument.
C
C When a valid file is found, the CAL_DEV column value also gets recorded to 
C the ONLINE argument.
C
C If no rows are found matching the input arguments then a STATUS of 1 is 
C returned.
C
C Arguments:   
C chatter       (i) : Chattiness flag
C cif           (i) : the path and name of the calibration index file.
C tele          (i) : the value of the TELESCOP column to be selected
C instr         (i) : the value of the INSTRUME column to be selected
C detnam        (i) : the value of the DETNAM column to be selected
C filt          (i) : the value of the FILTER column to be selected
C codenam       (i) : the value of the CAL_CNAM column to be selected
C startreftime  (i) : the value which is compared to the REF_TIME column
C stopreftime   (i) : the value which is compared to the REF_TIME column
C expr          (i) : boolean expression used to search CAL_CBD column
C maxret        (i) : the maximum number of entries to return 
C file          (r) : an array containing the system dependent path to the 
C                       calibration file
C extno         (r) : an array containing the extension numbers of the 
C                       calibration datasets found in the CAL_XNO column
C online        (r) : array specifying on-line/off-line status of files 
C                       in file array
C nfound        (r) : the total number of entries which met the selection 
C                       criteria
C status  (r) : success flag for this routine
C                            0 = everything is OK
C                            1 = no rows found matching input arguments
C                            2 = cannot open cif
C                            3 = problem reading CIF
C Origin:      
C  Written for the Calibration Database
C
C Authors/Modification History:
C  Ron Zellar (1.0.0:93 Apr 7), original version
C  Ron Zellar (2.0.0:94 May 31) Added detnam and filt arguments
C                               Enhanced if block selection so if detnam or 
C                               filt = '-', then not used in selection.
C  Ron Zellar (3.0.0:94 Jul 18) Enhanced if block selection so if cif 
C                               gives 'NONE' for DETNAM or FILTER then 
C                               comparison is TRUE.  Added expr parser and
C                               comparison.  Modified file and extno 
C                               arguments to be arrays. Added nfound argument
C                               and maxret argument, removed dev and dir 
C                               arguments.
C  Ron Zellar (3.1.0:94 Jul 20) RENAMED cifsel to cifsl
C  Ron Zellar (4.0.0:94 Aug  3) Added online argument
c  Ian George (5.0.0:96 Feb 01) Added CHATTER & STOPREFTIME as passed params
C                               Removed QUIET as passed, plus cosmetics
c  Ian George (5.0.1:96 Jun 03) Initialized tmp_sum to 0 on entry
c  Lorraine Breedon (5.0.2:97 Mar 01) Increased size of temporary array
c  Mike Tripicco (5.0.3:98 Jan 16) Changed nulval arg in ftgcvd from 0 to 0.0d0
c  Peter D Wilson(5.0.4:00 Jun 26) Suppress file-not-found error
C                                  when chatter<0
C  Mike Corcoran (5.0.5 06-Jun-20) compare CBD values to expr only for
C                               those rows that match televal, insval,
C                               detval,filtval, codval and qulval
C
C MFC (5.1 June 2011) - now determine mission matching from CALDB.CONFIG file;
C               remove matching mission string to TELESCOP in the 
C               caldb.indx file.  This makes it easier to handle mission
C               aliases/renaming (GLAST vs. Fermi, for example). 
C               This requires that all the mission data are stored in  
C               subdirectories of the main directory which stores the caldb.indx
C               file (for eg. all fermi data should be under
C               $CALDB/data/glast), as has been the convention
C               Added a warning if the TELESCOP value in the caldb.indx file
C               does not match the mission string if chatter >0
C                               
c        character(7) vers
c        parameter (vers = '5.1')

C---------------------------------------------------------------------
*-
c Internals
        character(9) subname
        parameter (subname = 'ero_cifsl')
        character(4) cval1, cval2,str4char
        character(10) telval, insval, filtval
        character(20) codval, devval, detval, caldbvar
        character(70) dirval
        character(30) sjunk
        character(160) contxt,
     &                errstr
        integer unit, errstat,ijunk,i,j,nax2val,qulval,extval
        integer telcol, inscol, codcol, refcol, qulcol, devcol, 
     &          dircol, filcol, extcol, tellen, inslen, codlen,
     &          fcstln, detcol, filtcol, detlen, filtlen,
     &          errstlen
        integer retcnt, cbdcol

        double precision refval, maxref
        logical ljunk,cbdres
        integer tmp_sum, tmp_max, num_valid_files
c       parameter (tmp_max = 100)
        parameter (tmp_max = 500)
        integer tmp_row(tmp_max)
        double precision tmp_tim(tmp_max)
        logical tmp_is_valid(tmp_max) 

C eRO variables
        integer :: envstat, ib1, ib2, ierr, len, is, is0
        integer :: relcol, instcol, typcol, fltcol, verscol, overcol 
        integer :: cvers, ip, iv, ivers, ie, svers, nfield, nfield1
        integer, dimension(tmp_max) :: tmp_vrs,tmp_ext
        integer, save :: nrow=0

        character(len=256) :: calversfile, oname, filename
        character(len=256), dimension(tmp_max) :: tmp_ovr,tmp_fnm
        character(len=20) :: versstring, release, task
        character(len=64) :: version
        character(10) startMJDstr,stopMJDstr

        logical, save :: first = .true., vfirst = .true.
        logical :: exist, found, cfound
        logical :: undef, passes_screen_1

        type :: vers_type
          character(len=20) :: release
          character(len=10) :: inst
          character(len=20) :: type
          character(len=10) :: filt
          integer(int16) :: version
          character(len=128) :: over
          integer :: ncomp 
        end type

        type (vers_type), dimension(:), allocatable, save :: calvers

        write(startMJDstr,'(f10.4)'), startreftime
        write(stopMJDstr, '(f10.4)'), stopreftime

C Set the name of the environment variable used in this routine
        caldbvar = 'CALDB'

C Initialize 
        tmp_sum = 0
        status  = 0
        errstat = 0
        devval  = ' '
        dirval  = ' '
        retcnt  = 0
        nfound  = 0
        errstlen = 0
        do j=1,maxret
             file(j) = ' '
        enddo
C eR0: len_trim will be 0 if no override file name read from calvers file
        oname=" "

c Give user info if requested
c        contxt = ' using '//subname//' '//vers
c        call wtinfo(chatter,20,1,contxt)

C eRO: read calvers.indx file, get version number
        exist=.false.
        found=.false. 
        call get_environment_variable("SASS_CALVERS", calversfile, len, envstat)
        if (envstat /= 0) then
          call wrerr(subname,1,"SASS_CALVERS not set")
        else
          inquire(file=calversfile, exist=exist)
        end if

        cvers = -1
        if (exist) then
        
          if (vfirst) then
            call ftgiou(unit,status)
            call fitserr("ftgiou", status, ierr)
            call fttopn(unit, calversfile, 0, status)
            call fitserr("fttopn", status, ierr)
            call ftgnrw(unit, nrow, status)
            call fitserr("ftgnrw", status, ierr)
            call ftgcno(unit, .true., "RELEASE", relcol, status)
            call fitserr("ftgcno", status, ierr)
            call ftgcno(unit, .true., "INSTRUME", instcol, status)
            call fitserr("ftgcno", status, ierr)
            call ftgcno(unit, .true., "CAL_CNAM", typcol, status)
            call fitserr("ftgcno", status, ierr)
            call ftgcno(unit, .true., "FILTER", fltcol, status)
            call fitserr("ftgcno", status, ierr)
            call ftgcno(unit, .true., "CALVERS", verscol, status)
            call fitserr("ftgcno", status, ierr)
            call ftgcno(unit, .true., "OVERRIDE", overcol, status)
            call fitserr("ftgcno", status, ierr)

            allocate (calvers(nrow))

            call ftgcvs(unit, relcol, 1, 1, nrow, " ", calvers%release, undef, status)
            call fitserr("ftgcvs", status, ierr)
            call ftgcvs(unit, instcol, 1, 1, nrow, " ", calvers%inst, undef, status)
            call fitserr("ftgcvs", status, ierr)
            call ftgcvs(unit, typcol, 1, 1, nrow, " ", calvers%type, undef, status)
            call fitserr("ftgcvs", status, ierr)
            call ftgcvs(unit, fltcol, 1, 1, nrow, " ", calvers%filt, undef, status)
            call fitserr("ftgcvs", status, ierr)
            call ftgcvi(unit, verscol, 1, 1, nrow, 0, calvers%version, undef, status)
            call fitserr("ftgcvi", status, ierr)
            call ftgcvs(unit, overcol, 1, 1, nrow, " ", calvers%over, undef, status)
            call fitserr("ftgcvs", status, ierr)

            call ftclos(unit) 
            call fitserr("ftclose", status, ierr)

            vfirst = .false.
          end if 

C eRO: get release name
          versstring = version()
          ib1=index(versstring," ")
          ib2=index(versstring(ib1+1:)," ")
          release=versstring(ib1+1:ib1+ib2-1)

C eRO: get version number and override file name from calvers.indx file
          cfound = .false.
          nfield1=0
          do i=1,nrow
            if (trim(release) == trim(calvers(i)%release)) then
C eRO: get number of non-empty fields in calvers file
              nfield=0
              if (len_trim(calvers(i)%inst)>0) nfield=nfield+4
              if (len_trim(calvers(i)%type)>0) nfield=nfield+2
              if (len_trim(calvers(i)%filt)>0) nfield=nfield+1
              if (nfield>nfield1) then
C eRO: overwrite cvers and oname if entry with more non-empty fields is found
                if ((len_trim(calvers(i)%inst)==0 .or. trim(instr) == trim(calvers(i)%inst)) .and.
     &              (len_trim(calvers(i)%type)==0 .or. trim(codenam) == trim(calvers(i)%type)) .and.
     &              (len_trim(calvers(i)%filt)==0 .or. trim(filt) == trim(calvers(i)%filt))) then
                  nfield1 = nfield
                  cfound = .true.
                  if (cvers < calvers(i)%version) then
                    cvers = calvers(i)%version
                    oname = calvers(i)%over
                  end if
                end if
              end if
            end if
          end do
          if (.not. cfound) call wrchatty(subname,
     &      trim(instr)//" "//trim(codenam)//" "//trim(filt)//
     &      " not found in calvers file, so I'll use the highest"//
     &      " matching version in the CIF.")

c          deallocate (calvers)
        else
          call wrlog(subname,"SASS_CALVERS file not found, all info will be sought from the CIF.")
        end if 

C open the CIF
        call cgetlun(unit)
        call ftopen(unit, cif, 0, ijunk, errstat)
        if(errstat.ne.0)then
c          contxt = 'Unable to open CIF'
c          if(chatter.ge.0) call wrerr(subname, 3, contxt)
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to open CIF file "//cif(:fcstln(cif)))
          status = 2
          goto 999
        endif

C move to the first extension
        call ftmahd(unit, 2, ijunk, errstat)
        if(errstat.ne.0)then
c          contxt = 'Problem reading CIF'
c          call wrerr(subname, 3, contxt)
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Problem reading CIF file "//cif(:fcstln(cif)))
          status = 2
          goto 999
        endif

C find out how many rows the CIF contains
        call ftgkyj(unit, 'NAXIS2', nax2val, sjunk, errstat)
        if(errstat.ne.0)then
c          contxt = 'Problem reading NAXIS2 of CIF'
c          call wrerr(subname, 3, contxt)
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Problem reading NAXIS2 of CIF file.")
          status = 2
          goto 999
        endif
        if(nax2val.eq.0)then
c          contxt = 'CIF appears to be empty'
c          call wterrm(subname, vers, contxt) 
c          call wrerr(subname, 3, contxt)
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "CIF file appears to be empty.")
          status = 1
          goto 999
        endif
        
C get the column numbers of all the columns to be read

        call ftgcno(unit,.true.,'TELESCOP', telcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find TELESCOP column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find TELESCOP column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'INSTRUME', inscol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find INSTRUME column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find INSTRUME column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'DETNAM',   detcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname, 3, "Unable to find DETNAM column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find DETNAM column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'FILTER',filtcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname, 3, "Unable to find FILTER column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find FILTER column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_CNAM', codcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_CNAM column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_CNAM column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'REF_TIME', refcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find REF_TIME column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find REF_TIME column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_QUAL', qulcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_QUAL column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_QUAL column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_DEV',  devcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname, 3, "Unable to find CAL_DEV column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_DEV column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_DIR',  dircol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_DIR column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_DIR column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_FILE', filcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_FILE column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_FILE column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_XNO',  extcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_XNO column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_XNO column in CIF file.")
          status = 2
          goto 999
        endif

        call ftgcno(unit,.true.,'CAL_CBD',  cbdcol,errstat)
        if(errstat.ne.0)then
c          call wrerr(subname,3,"Unable to find CAL_CBD column")
c          contxt = 'offending file: '//cif(:fcstln(cif))
c          call wtinfo(chatter,1,1,contxt)
          call wrerr(subname, 3, "Unable to find CAL_CBD column in CIF file.")
          status = 2
          goto 999
        endif

C read each row and check against input arguments.  If row matches input 
C arguments then read returned values and write them to returned arguments.
        do 100 i=1, nax2val
          call ftgcvs(unit,telcol,i,1,1,' ',telval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF file "//cif(:fcstln(cif)))
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')telcol
c            contxt='problematic row = '//cval1//', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

          call ftgcvs(unit,inscol,i,1,1,' ',insval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')inscol
c            contxt='problematic row = '//cval1//', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C ... Since DETNAM is not required to make a unique selection it is removed
c     from the selection criteria by making all detval values equal to the
c     input value (a space '-').

          if (detnam .eq. '-') then
            detval = '-'
          else
            call ftgcvs(unit,detcol,i,1,1,' ',detval,
     &                 ljunk,errstat)
            if (errstat .ne. 0) then
c              call wrerr(subname, 3, "Error reading CIF")
c              contxt = 'filename: '//cif(:fcstln(cif))
c              call wtinfo(chatter,1,1,contxt)
              write(cval1,'(I4)')i
              write(cval2,'(I4)')detcol
c              contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c              call wtinfo(chatter,1,2,contxt)
              call wrerr(subname, 3, "Error reading CIF file "//
     &          cif(:fcstln(cif))//". Problematic row="//
     &          trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
              status = 2
              goto 999
            endif
          endif

C       As for DETNAM, so it is for FILTER...

          if (filt .eq. '-' ) then
            filtval = '-'
          else
            call ftgcvs(unit,filtcol,i,1,1,' ',filtval,
     &                ljunk,errstat)
            if (errstat .ne. 0) then
c              call wrerr(subname, 3, "Error reading CIF")
c              contxt = 'filename: '//cif(:fcstln(cif))
c              call wtinfo(chatter,1,1,contxt)
              write(cval1,'(I4)')i
              write(cval2,'(I4)')filtcol
c              contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c              call wtinfo(chatter,1,2,contxt)
              call wrerr(subname, 3, "Error reading CIF file "//
     &          cif(:fcstln(cif))//". Problematic row="//
     &          trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
              status = 2
              goto 999
            endif
          endif

          call ftgcvs(unit,codcol,i,1,1,' ',codval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')codcol
c            contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

          call ftgcvd(unit,refcol,i,1,1,0.0d0,refval,ljunk,errstat)
c     changed 0 to 0.0d0 in line above (mjt/16jan98)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')refcol
c            contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

          call ftgcvj(unit,qulcol,i,1,1,0,qulval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')qulcol
c            contxt='problematic row = '//cval1//
c     &                  ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C eRO: get filename from cal index file, read version number
          call ftgcvs(unit,filcol,i,1,1,' ',filename,ljunk,errstat)
          call fitserr("ftgcvs", status, ierr)
          ip=index(filename,".")
          iv=index(filename(ip-5:),"v")
          read(filename(ip-5+iv:ip-1),*) ivers

C eRO: read extension.
          call ftgcvj(unit,extcol,i,1,1,0,extval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')i
            write(cval2,'(I4)')extcol
c            contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C get the lengths of the strings read in
          tellen = fcstln(telval)
          inslen = fcstln(insval)
          codlen = fcstln(codval)
          filtlen= fcstln(filtval)
          detlen = fcstln(detval)

C MFC Addition 201106 version 5.1
C
C Prior to version 5.1: Because missions are often renamed 
C after launch, trying to match the TELESCOP name
C in the caldb.indx file will result in a mis-match if the post-launch mission 
C name is used (for eq. searching for Fermi data won't match files where
C TELESCOP='GLAST    '
C
C Version 5.1 change: The caldb for a given mission 
C is stored in bcf and cpf subdirectories
C of the directory where the caldb.indx file is located, so that the 
C mission data is effectively determined by the location of the caldb.indx file.
C So we REQUIRE that the specified mission name match the name of the directory 
C in which the caldb.indx is stored (as has been the convention)
C and therefore we don't need to require that the mission name string equal the 
C value of the TELESCOP keyword stored in the CALDB.
C However we should compare the mission string to the TELESCOP value and 
C warn the user of mismatches (which could indicate a problem in the 
C caldb.indx file)
C 
C  MFC
          if(telval(:tellen).ne.tele) then 
            errstr='Mission name '//tele(:fcstln(tele))//' Does not'
     &          //' Match TELESCOP= '//telval(:tellen)
            errstr=errstr//telval(:tellen)
            contxt=errstr(:fcstln(errstr)) 
c            call wtwarm(subname, version, chatter, 1, contxt)
            call wrerr(subname, 1, contxt)
          endif
C  MFC          if((telval(:tellen).eq.tele).and.
C  the read values match the input values read in the returned values
          passes_screen_1 = .true. ! default

          if(insval(:inslen).ne.instr)
     &      passes_screen_1 = .false.
          if( (detval(:detlen).ne.detnam) .and. (detval(:detlen).ne.'NONE') )
     &      passes_screen_1 = .false.
          if( (filtval(:filtlen).ne.filt) .and. (filtval(:filtlen).ne.'NONE') )
     &      passes_screen_1 = .false.
          if(codval(:codlen).ne.codenam)
     &      passes_screen_1 = .false.
          if(qulval.ne.0)
     &      passes_screen_1 = .false.
C eRO: check version number if found in calvers file
          if(cvers .ge. 0 .and. ivers .ne. cvers)
     &      passes_screen_1 = .false.

          if(passes_screen_1) then
C Check to see if this row is valid for expr
            if (expr.eq.'-') then
              cbdres = .true.
            else
              call evexp2(chatter, unit,i,cbdcol,expr,
     &                          cbdres,errstat)
              if (errstat .ne. 0) then
c                call wrerr(subname, 3, "Error reading CIF")
c                contxt = 'filename: '//cif(:fcstln(cif))
c                call wtinfo(chatter,1,1,contxt)
                write(cval1,'(I4)')i
                write(cval2,'(I4)')cbdcol
c                contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c                call wtinfo(chatter,1,2,contxt)
                call wrerr(subname, 3, "Error reading CIF file "//
     &            cif(:fcstln(cif))//". Problematic row="//
     &            trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
                status = 2
                goto 999
              endif 
            endif 

            if (cbdres) then 

C ..... OK, this means we've found a calibration dataset of the requested 
C       type which satisfies all the limitations EXCEPT perhaps the date/time
C       Lob these into temporary arrays
              tmp_sum = tmp_sum + 1
              if(tmp_sum.gt.tmp_max) then
c                call wtwarm(subname,version,chatter,1,
c     &            'Internal Array exceeded (param tmp_max)')
c                call wtinfo(chatter,1,1,
c     &            'Some files may have been skipped')
c                contxt = 'Suggest more conservative '//
c     &            'selection criteria'
c                call wtinfo(chatter,1,2,contxt)
                call wrerr(subname, 1, "Internal array exceeded (param tmp_max)."//
     &            " Some files may have been skipped. Suggest you use more"//
     &            " conservative selection criteria.")
                goto 669
              endif ! tmp_sum if
              tmp_row(tmp_sum) = i
              tmp_tim(tmp_sum) = refval
              tmp_is_valid(tmp_sum) = .true. ! default.
C eRO: save version, override filename, CIf filename and extension.
              tmp_vrs(tmp_sum) = ivers
              tmp_ovr(tmp_sum) = oname
              tmp_fnm(tmp_sum) = filename
              tmp_ext(tmp_sum) = extval
            endif 
          endif ! passes_screen_1
100     continue ! end loop over CIF rows.
669     continue

        if(tmp_sum<1) then
          call wrerr(subname, 2, "No files match instrument,"//
     &      "filter etc.")
          status = 2
          goto 999
        end if

        write(str4char,'(i4)') tmp_sum
        call wrchatty(subname, trim(adjustl(str4char))//
     &    " files matched instrument, filter etc.")

        write(str4char,'(i4)') num_valid_files
        call wrchatty(subname, "Testing against start MJD "//
     &    trim(startMJDstr)//" and stop MJD "//trim(stopMJDstr))

C Now loop through the temp stuff twice, once finding the latest 
C validity start date/time BEFORE startreftime, then flagging the 
C earlier (unwanted) datasets to be ignored.
C If startreftime = -99, then we have no constraints in early files 
C and want all files starting at the year dot up until stopreftime
        maxref = 0.
        if(startreftime.EQ.-99) then
          goto 666
        else
          do i = 1, tmp_sum
            if((tmp_tim(i).le.startreftime).and.
     &           (tmp_tim(i).gt.maxref)) maxref =  tmp_tim(i)
          enddo
          do i = 1, tmp_sum
            if(tmp_tim(i).lt.maxref) tmp_is_valid(i) = .false.
          enddo
        endif

666     continue

c Now loop through flagging all datasets after the stopreftime to
c be ignored.
C If stopreftime = -99, then we have no constraints on late files
C and want all files after startreftime
        if(stopreftime.EQ.-99) then
          goto 667
        else
          do i = 1, tmp_sum
            if(tmp_tim(i).gt.stopreftime) tmp_is_valid(i) = .false.
          enddo
        endif

667     continue

        num_valid_files = 0
        do i = 1, tmp_sum
          if(tmp_is_valid(i)) num_valid_files = num_valid_files + 1
        enddo
        if(num_valid_files<1) then
          call wrerr(subname, 2, "No files match the time criteria.")
          status = 2
          goto 999
        end if

        write(str4char,'(i4)') num_valid_files
        call wrchatty(subname, trim(adjustl(str4char))//" files matched the time criteria.")

        if(cvers<0) then
c eRO:    No matching file version was found in $SASS_CALVERS, so keep only the
c         highest version number of any subset of files having the same REF_TIME and CAL_XNO.

          do i = 1, tmp_sum
            if(.not.tmp_is_valid(i)) goto 670

            do j = 1, tmp_sum
              if(i.eq.j .or. .not.tmp_is_valid(j)) goto 668

              if(tmp_tim(i).eq.tmp_tim(j) .and. tmp_ext(i).eq.tmp_ext(j)) then
                if(tmp_vrs(i).eq.tmp_vrs(j)) then
                  call wrerr(subname, 3, trim(tmp_fnm(i))//
     &              " and "//trim(tmp_fnm(j))//" have identical REF_TIME")
                  status = 2
                  goto 999
                else if(tmp_vrs(i).lt.tmp_vrs(j)) then
                  tmp_is_valid(i) = .false.
                else !  tmp_vrs(i).gt.tmp_vrs(j)
                  tmp_is_valid(j) = .false.
                end if
              end if
668         enddo
670       enddo

          num_valid_files = 0
          do i = 1, tmp_sum
            if(tmp_is_valid(i)) num_valid_files = num_valid_files + 1
          enddo
          if(num_valid_files<1) then
            call wrerr(subname, 2, "No files match the version criteria.")
            status = 2
            goto 999
          end if
        end if ! no version found in $SASS_CALVERS.

c So now we're ready for the final pass through the CIF, reading all
c the other keywords etc for the datasets within the requested time
c window.
        
        do i = 1, tmp_sum
          if(.NOT.tmp_is_valid(i)) goto 700
          
          nfound = nfound + 1
          if (retcnt.ge.maxret) then
            call wrerr(subname, 1, "More files have been found than the user's specified maximum number.")
            goto 700
          end if
          retcnt = retcnt + 1

C ........ Get directory value
          call ftgcvs(unit,dircol,tmp_row(i),1,1,' ',
     &           dirval,ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')tmp_row(i)
            write(cval2,'(I4)')dircol
c            contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C ........ Get filename value
          call ftgcvs(unit,filcol,tmp_row(i),1,1,' ',
     &             file(retcnt),ljunk,errstat)
          if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')tmp_row(i)
            write(cval2,'(I4)')filcol
c            contxt='problematic row = '//cval1//
c     &                          ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C ........ construct system dependent path
          call cpthnm(caldbvar,dirval,file(retcnt),errstat)
          if (errstat .ne. 0) then
c            call wtferr(subname, vers, errstat,
c     &                          'Error constructing path')
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')tmp_row(i)
c            write(cval2,'(I4)')detcol
c            contxt='problematic row = '//cval1
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error constructing path. File="//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1)))
            status = 2
            goto 999
          endif
                    
C ........ Get extension number 
c eRO: We are reading and storing the extension number already, so just copy it over.
c         call ftgcvj(unit,extcol,tmp_row(i),1,1,0,
c     &             extno(retcnt),ljunk,errstat)
c         if (errstat .ne. 0) then
c            call wrerr(subname, 3, "Error reading CIF")
c           contxt = 'filename: '//cif(:fcstln(cif))
c           call wtinfo(chatter,1,1,contxt)
c           write(cval1,'(I4)')tmp_row(i)
c           write(cval2,'(I4)')extcol
c           contxt='problematic row = '//cval1//
c     &                         ', col = '//cval2
c           call wtinfo(chatter,1,2,contxt)
c            status = 2
c            goto 999
c         endif
          extno(retcnt) = tmp_ext(i)

C ........ Get online/offline value
          call ftgcvs(unit,devcol,tmp_row(i),1,1,' ',
     &                 online(retcnt),ljunk,errstat)
          if (errstat .ne. 0) then
            call wrerr(subname, 3, "Error reading CIF")
c            contxt = 'filename: '//cif(:fcstln(cif))
c            call wtinfo(chatter,1,1,contxt)
            write(cval1,'(I4)')tmp_row(i)
            write(cval2,'(I4)')devcol
c            contxt='problematic row = '//cval1//
c     &                  ', col = '//cval2
c            call wtinfo(chatter,1,2,contxt)
            call wrerr(subname, 3, "Error reading CIF file "//
     &        cif(:fcstln(cif))//". Problematic row="//
     &        trim(adjustl(cval1))//", col="//trim(adjustl(cval2)))
            status = 2
            goto 999
          endif

C eRO: override filename and extention number if override name is present
          if (len_trim(tmp_ovr(i)) > 0) then
            ie = index(tmp_ovr(i),"+")
            if (ie>0) then
              file(retcnt) = tmp_ovr(i)(1:ie-1)
              read(tmp_ovr(i)(ie+1:),*) extno(retcnt)
            else
              file(retcnt) = tmp_ovr(i)
            end if
          end if 
    
700     enddo

C if maxref is zero then no datasets matched the input values
C        if(maxref.eq.0)status=1

C Close the Index File
        errstat = 0
        call ftclos(unit,errstat)
        if (errstat .ne. 0) then
          call wrerr(subname, 3, "Problem closing CIF")
          status = 7
          goto 999
        endif

C Free up the logical unit number I was using
        call cfrelun(unit)

        if (nfound<=0) call wrerr(subname, 2,
     &              "No dataset found matching the selection criteria")

C eRO: copy selected cal file names to log

C eRO: append cal file name to log and history keyword
        do i=1,retcnt
          is=1;is0=0
          do while (is>0)
            is0=is+is0
            is=index(file(1)(is0:),"/")
          end do
          if (first) then 
            sass_history = trim(sass_history) //  " CALIB:" //
     &                     trim(file(i)(is0:))

            call wrlog(subname, "Accessing cal file " // 
     &                                     trim(file(i)(is0:)))
            first = .false. 
          else
            if(index(sass_history,trim(file(i)(is0:)))==0)then
              sass_history = trim(sass_history) // " " // 
     &                                 trim(file(i)(is0:))
              call wrlog(subname, "Accessing cal file " // 
     &                                       trim(file(i)(is0:)))
            end if
          end if
        end do

c Final error checking
999     if ((errstat.ne.0).or.(status.gt.0)) then
c          call wrerr(subname, 4, "Fatal - aborting")
        endif

        return
        end
